@@ -8,6 +8,7 @@ import sys
 import termios
 import time
 import tty
+import warnings
 
 import paramiko
 
@@ -29,7 +30,8 @@ class bcolors:
     UNDERLINE = '\033[4m'
 
 
-def remote_command(client, command, pty=False, print_command=False):
+def remote_command(client, command, pty=False, dir=None, source_bashrc=False,
+                   print_command=False):
     transport = client.get_transport()
     session = transport.open_session()
 
@@ -37,6 +39,12 @@ def remote_command(client, command, pty=False, print_command=False):
         session.setblocking(0)
         session.get_pty()
 
+    if dir is not None:
+        command = f'cd {dir}; {command}'
+
+    if source_bashrc:
+        command = f'source $HOME/.bashrc; {command}'
+
     session.exec_command(command)
 
     if print_command:
@@ -75,10 +83,18 @@ def watch_command(command, stop_condition=None, keyboard_int=None,
                 return False
         return not stop_condition()
 
+    stop = False
     try:
-        while continue_running():
+        while not stop:
             time.sleep(0.01)
 
+            # We consume the output one more time after it's done.
+            # This prevents us from missing the last bytes.
+            if (timeout is not None) and (time.time() > deadline):
+                stop = True
+            elif not continue_running():
+                stop = True
+
             if command.recv_ready():
                 data = command.recv(512)
                 decoded_data = data.decode('utf-8')
@@ -95,8 +111,6 @@ def watch_command(command, stop_condition=None, keyboard_int=None,
                     sys.stderr.write(decoded_data)
                     sys.stderr.flush()
 
-            if (timeout is not None) and (time.time() > deadline):
-                break
     except KeyboardInterrupt:
         if keyboard_int is not None:
             keyboard_int()
@@ -152,7 +166,8 @@ def get_ssh_client(host, nb_retries=0, retry_interval=1):
     return client
 
 
-def run_console_commands(console, commands, timeout=1, console_pattern=None):
+def run_console_commands(console, commands, timeout=1, console_pattern=None,
+                         verbose=False):
     if not isinstance(commands, list):
         commands = [commands]
 
@@ -167,7 +182,8 @@ def run_console_commands(console, commands, timeout=1, console_pattern=None):
         output += watch_command(console,
                                 keyboard_int=lambda: console.send('\x03'),
                                 timeout=timeout, stop_pattern=console_pattern,
-                                max_match_length=console_pattern_len)
+                                max_match_length=console_pattern_len,
+                                stdout=verbose, stderr=verbose)
 
     return output
 
@@ -208,18 +224,20 @@ def posix_shell(chan):
 
 class RemoteIntelFpga:
     def __init__(self, host: str, fpga_id: str, remote_dir: str,
-                 load_bitstream: bool = True):
+                 load_bitstream: bool = True, verbose=False):
         self.host = host
         self.fpga_id = fpga_id
         self._ssh_client = None
         self.jtag_console = None
         self.remote_dir = remote_dir
+        self.verbose = verbose
 
         self.setup(load_bitstream)
 
     def run_jtag_commands(self, commands):
         return run_console_commands(self.jtag_console, commands,
-                                    console_pattern='\r\n% ')
+                                    console_pattern='\r\n% ',
+                                    verbose=self.verbose)
 
     def launch_console(self, max_retries=5):
         retries = 0
@@ -231,11 +249,12 @@ class RemoteIntelFpga:
             app = remote_command(self.ssh_client, cmd, pty=True, dir=cmd_path,
                                  source_bashrc=True)
             watch_command(app, keyboard_int=lambda: app.send('\x03'),
-                          timeout=10)
+                          timeout=10, stdout=self.verbose, stderr=self.verbose)
 
             app.send('source path.tcl\n')
             output = watch_command(app, keyboard_int=lambda: app.send('\x03'),
-                                   timeout=2)
+                                   timeout=2, stdout=self.verbose,
+                                   stderr=self.verbose)
             lines = output.split('\n')
             lines = [
                 ln for ln in lines
@@ -268,7 +287,8 @@ class RemoteIntelFpga:
         while load_bitstream:
             app = remote_command(self.ssh_client, cmd, pty=True, dir=cmd_path,
                                  source_bashrc=True)
-            output = watch_command(app, keyboard_int=lambda: app.send('\x03'))
+            output = watch_command(app, keyboard_int=lambda: app.send('\x03'),
+                                   stdout=self.verbose, stderr=self.verbose)
             status = app.recv_exit_status()
             if status == 0:
                 break
@@ -277,6 +297,8 @@ class RemoteIntelFpga:
                 raise RuntimeError('Synchronization failed, try power cycling '
                                    'the host')
 
+            warnings.warn('Failed to load bitstream, retrying.')
+
             retries += 1
             if retries >= 5:
                 raise RuntimeError(f'Failed to load bitstream {retries} times')
